# Unit constants
import math
from . import clibrebound
from ctypes import c_char_p, c_uint32

def hash_to_unit(hash):
    clibrebound.reb_hash.restype = c_uint32
    for u in times_SI.keys():
        uhash = clibrebound.reb_hash(c_char_p(u.encode("ascii")))
        if uhash == hash:
            return u
    for u in masses_SI.keys():
        uhash = clibrebound.reb_hash(c_char_p(u.encode("ascii")))
        if uhash == hash:
            return u
    for u in lengths_SI.keys():
        uhash = clibrebound.reb_hash(c_char_p(u.encode("ascii")))
        if uhash == hash:
            return u
    return None

# All units entered in SI (kg, m, s)
G_SI = 6.67408e-11
times_SI = {'s':1.,
    'hr':3600.,
    'day': 86400.,
    'days': 86400.,
    'd': 86400.,
    'yr':31557600., # Julian year (exact)
    'yrs':31557600., 
    'jyr':31557600.,
    'sidereal_yr':31558149.7635,
    'yr2pi':math.sqrt(149597870700.**3/1.3271244004193938e20), # chosen to make G=1
    'kyr':31557600.*1.e3,
    'myr':31557600.*1.e6,
    'gyr':31557600.*1.e9}
lengths_SI =  {'m':1.,
    'cm':0.01,
    'km':1000.,
    'au':149597870700.,
    'aus':149597870700.,
    'pc':3.085677581e16,
    'parsec':3.085677581e16
    }

    #What we measure accurately is GM, so set mass units such that G*M gives the value of GM in horizons.py (in the list at the end of horizons.py, the NAIF codes ending in 99 refer to the planets, single digits to the total mass of the planet plus its moons).  Have to multiply by 10**9 since that list has G in kg^-1km^3/s^2 and we use SI.

masses_SI = {'kg':1.,
    'g':1.0e-3,
    'gram':1.0e-3,
    'msun':1.3271244004193938E+11/G_SI*10**9,
    'solarmass':1.3271244004193938E+11/G_SI*10**9,
    'sunmass':1.3271244004193938E+11/G_SI*10**9,
    'msolar':1.3271244004193938E+11/G_SI*10**9,
    'mmercury':2.2031780000000021E+04/G_SI*10**9,
    'mvenus':3.2485859200000006E+05/G_SI*10**9,
    'mearth':3.9860043543609598E+05/G_SI*10**9,
    'mmars':4.282837362069909E+04/G_SI*10**9,
    'mjupiter':1.266865349218008E+08/G_SI*10**9,
    'msaturn':3.793120749865224E+07/G_SI*10**9,
    'muranus':5.793951322279009E+06/G_SI*10**9,
    'mneptune':6.835099502439672E+06/G_SI*10**9,
    'mpluto':8.696138177608748E+02/G_SI*10**9}


def units_convert_particle(p, old_l, old_t, old_m, new_l, new_t, new_m):
    p.m = convert_mass(p.m, old_m, new_m)
    p.x = convert_length(p.x, old_l, new_l) 
    p.y = convert_length(p.y, old_l, new_l)
    p.z = convert_length(p.z, old_l, new_l)
    p.vx = convert_vel(p.vx, old_l, old_t, new_l, new_t)
    p.vy = convert_vel(p.vy, old_l, old_t, new_l, new_t)
    p.vz = convert_vel(p.vz, old_l, old_t, new_l, new_t)
    p.ax = convert_acc(p.ax, old_l, old_t, new_l, new_t)
    p.ay = convert_acc(p.ay, old_l, old_t, new_l, new_t)
    p.az = convert_acc(p.az, old_l, old_t, new_l, new_t)
    return p

def convert_mass(mass, old_m, new_m):
    return mass*masses_SI[old_m]/masses_SI[new_m]

def convert_length(length, old_l, new_l):
    return length*lengths_SI[old_l]/lengths_SI[new_l]

def convert_vel(vel, old_l, old_t, new_l, new_t):
    in_SI=vel*lengths_SI[old_l]/times_SI[old_t]
    return in_SI*times_SI[new_t]/lengths_SI[new_l]

def convert_acc(acc, old_l, old_t, new_l, new_t):
    in_SI=acc*lengths_SI[old_l]/times_SI[old_t]**2
    return in_SI*times_SI[new_t]**2/lengths_SI[new_l]

def convert_G(newunits):
    new_l, new_t, new_m = newunits
    return G_SI*masses_SI[new_m]*times_SI[new_t]**2/lengths_SI[new_l]**3
       
def check_units(newunits):   
    if len(newunits) != 3:
        raise Exception("Error: Need to pass exactly 3 units for length, time, and mass (any order), see ipython_examples/Units.ipynb")

    if isinstance(newunits, dict): 
        # keys are not important as they are inferred from the values anyway
        newunits = newunits.values() 
    
    l_unit = t_unit = m_unit = None
    for unit in newunits:
        unit = unit.lower()
        if unit in lengths_SI:
            l_unit = unit
        if unit in times_SI:
            t_unit = unit
        if unit in masses_SI:
            m_unit = unit

    if l_unit is None or t_unit is None or m_unit is None:
        raise Exception("Error: Need to assign rebound.units a tuple consisting of 3 units for length, time, and mass (any order).  See ipython/examples/Units.ipynb.  If you passed such a tuple, at least one of your units isn't in our list.  Please update the dictionaries at the top of rebound/rebound/units.py and send a pull request!")

    return (l_unit, t_unit, m_unit)



from ctypes import Structure, c_double, POINTER, c_float, c_int, c_uint, c_uint32, c_int64, c_long, c_ulong, c_ulonglong, c_void_p, c_char_p, CFUNCTYPE, byref, create_string_buffer, addressof, pointer, cast
from .simulation import Simulation, BINARY_WARNINGS
from . import clibrebound 
import os
import sys
import math
import warnings

POINTER_REB_SIM = POINTER(Simulation) 

class SimulationArchive(Structure):
    """
    SimulationArchive Class.

    The SimulationArchive is a binary file format which includes all
    settings, constants as well as particle positions and velocities.
    This makes it possible to reproduce a simulation exactly 
    (down to the last bit). The SimulationArchive allows you to add
    an arbitrary number of snapshots. Simulations can be reconstructed
    from these snapshots. Since version 2 of the SimulationArchive
    (Spring 2018), you can change anything in-between snapshots,
    including settings like the integrator, the timestep, the number of
    particles. The file format is efficient in that only data
    that changed is stored in the SimulationArchive file. This is all
    done automatically. All the user has to do is call the function
    to create a snapshot.
    The SimulationArchive thus allows for fast access to any long-running 
    simulations. For a full discussion of the functionality see the paper 
    by Rein & Tamayo 2017.

    Requirements
    ------------
    When using the SimulationArchive, the user is responsible for 
    setting up any additional forces or post-timestep modifications that 
    were  present during the original integration. 
        
    Examples
    --------
    Here is a simple example:

    >>> sa = rebound.SimulationArchive("archive.bin")
    >>> sim = sa.getSimulation(t=1e6)
    >>> print(sim.particles[1])
    >>> for sim in sa:
    >>>     print(sim.t, sim.particles[1].e)

    """
    _fields_ = [("_inf", c_void_p),
                ("_filename", c_char_p),
                ("version", c_int), 
                ("size_first", c_long), 
                ("size_snapshot", c_long), 
                ("auto_interval", c_double), 
                ("auto_walltime", c_double), 
                ("auto_step", c_ulonglong), 
                ("nblobs", c_long), 
                ("offset", POINTER(c_uint32)), 
                ("t", POINTER(c_double)) 
                ]
    def __repr__(self):
        return '<{0}.{1} object at {2}, nblobs={3}>'.format(self.__module__, type(self).__name__, hex(id(self)), self.nblobs)

    def __init__(self,filename,setup=None, setup_args=(), process_warnings=True, reuse_index=None):
        """
        Arguments
        ---------
        filename : str
            Filename of the SimulationArchive file to be opened.
        setup : function
            Function to be called everytime a simulation object is created
            In this function, the user can setup additional forces
        setup_args : list
            Arguments passed to setup function.
        process_warnings : Bool
            Display warning messages if True (default). Only fail on major errors if set to False.
        reuse_index : SimulationArchive
            Useful when loading many large SimulationArchives. After loading the first 
            SimulationArchive, pass it as this argument when opening other SimulationArchives with the 
            same shape. Note: SimulationArchive shape must be exactly the same to avoid unexpected
            behaviour.

        """
        self.setup = setup
        self.setup_args = setup_args
        self.process_warnings = process_warnings
        w = c_int(0)
        if reuse_index:
            # Optimized loading
            clibrebound.reb_read_simulationarchive_with_messages(byref(self),c_char_p(filename.encode("ascii")), byref(reuse_index), byref(w))
        else:
            clibrebound.reb_read_simulationarchive_with_messages(byref(self),c_char_p(filename.encode("ascii")), None, byref(w))
        for majorerror, value, message in BINARY_WARNINGS:
            if w.value & value:
                if majorerror:
                    raise RuntimeError(message)
                else:  
                    # Just a warning
                    if process_warnings:
                        warnings.warn(message, RuntimeWarning)
        else:
            # Store for later
            self.warnings = w
        if self.nblobs<1:
            RuntimeError("Something went wrong. SimulationArchive is empty.")
        self.tmin = self.t[0]
        self.tmax = self.t[self.nblobs-1]

    def __del__(self):
        if self._b_needsfree_ == 1: 
            clibrebound.reb_free_simulationarchive_pointers(byref(self))

    def __str__(self):
        """
        Returns a string with details of this simulation archive.
        """
        return "<rebound.SimulationArchive instance, snapshots={0} >".format(str(len(self)))

    def __getitem__(self, key):
        PY3 = sys.version_info[0] == 3
        if PY3:
            int_types = int,
        else:
            int_types = int, long, 

        if isinstance(key, slice):
            raise AttributeError("Slicing not supported due to optimizations.")
        if not isinstance(key, int_types):
            raise AttributeError("Must access individual simulations with integer index.")
        if key < 0:
            key += len(self)
        if key>= len(self) or key<0:
            raise IndexError("Index out of range, number of snapshots stored in binary: %d."%len(self))
        
        w = c_int(0)
        sim = Simulation()
        clibrebound.reb_create_simulation_from_simulationarchive_with_messages(byref(sim), byref(self), c_long(key), byref(w))
        if self.setup:
            self.setup(sim, *self.setup_args)
        for majorerror, value, message in BINARY_WARNINGS:
            if w.value & value:
                if majorerror:
                    raise RuntimeError(message)
                else:  
                    # Just a warning
                    if self.process_warnings:
                        warnings.warn(message, RuntimeWarning)
        if sim.ri_eos.is_synchronized==0 or sim.ri_mercurius.is_synchronized==0 or sim.ri_whfast.is_synchronized==0 or sim.ri_mercurius.is_synchronized==0:
            warnings.warn("The simulation might not be synchronized. You can manually synchronize it by calling sim.integrator_synchronize().", RuntimeWarning)

        return sim
    
    def __setitem__(self, key, value):
        raise AttributeError("Cannot modify SimulationArchive.")

    def __delitem__(self, key):
        raise AttributeError("Cannot modify SimulationArchive.")

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        return self.nblobs  # number of SA snapshots (also counting binary at t=0)

    def _getSnapshotIndex(self, t):
        """
        Return the index for the snapshot just before t
        """
        if t>self.tmax or t<self.tmin:
            raise ValueError("Requested time outside of baseline stored in binary file.")
        # Bisection method
        l = 0
        r = len(self)
        while True:
            bi = l+(r-l)//2
            if self.t[bi]>t:
                r = bi
            else:
                l = bi
            if r-1<=l:
                bi = l
                break
        return bi, self.t[bi]

    def getSimulation(self, t, mode='snapshot', keep_unsynchronized=1):
        """
        This function returns a simulation object at (or close to) the requested time `t`. 
        Everytime this function is called a new simulation object is created.


        Arguments
        ---------
        t : float
            Requested time. Needs to be within tmin and tmax of this Simulation Archive.
        mode : str
            This argument determines how close the simulation should be to the requested time.
            There are three options. 
            - 'snapshot' This loads a nearby snapshot such that sim.t<t. This is the default.
            - 'close' This integrates the simulation to get to the time t but may overshoot by at most one timestep sim.dt.
            - 'exact' This integrates the simulation to exactly time t. This is not compatible with keep_unsynchronized=1. 
        keep_unsynchronized : int
            By default this argument is 1. This means that if the simulation had to be synchronized to generate this output, then it will nevertheless use the unsynchronized values if one integrates the simulation further in time. This is important for exact (bit-by-bit) reproducibility. If the value of this argument is 0, then one can modify the particles coordinates and these changes are taken into account when integrating the simulation further in time.
        
        Returns
        ------- 
        A rebound.Simulation object. Everytime the function gets called
        a new object gets created. 
        
        Examples
        --------
        Here is a simple example on how to load a simulation from a 
        Simulation Archive file with the `getSimulation` method.
        As the `mode` argument is set to `close`, the simulation
        will be integrated from the nearest snapshot to the request time.

        >>> sa = rebound.SimulationArchive("archive.bin")
        >>> sim = sa.getSimulation(t=1e6, mode="close")
        >>> print(sim.particles[1])
        >>> for sim in sa:
        >>>     print(sim.t, sim.particles[1].e)

        """
        if mode not in ['snapshot', 'close', 'exact']:
            raise AttributeError("Unknown mode.")

        bi, bt = self._getSnapshotIndex(t)
        sim = Simulation()
        w = c_int(0)
        clibrebound.reb_create_simulation_from_simulationarchive_with_messages(byref(sim),byref(self),bi,byref(w))

        # Restore function pointers and any additional setup required by the user provided functions
        if self.setup:
            self.setup(sim, *self.setup_args)

        if mode=='snapshot':
            if (sim.integrator=="mercurius" and sim.ri_mercurius.safe_mode == 1) or (sim.integrator=="whfast" and sim.ri_whfast.safe_mode == 1) or (sim.integrator=="saba" and sim.ri_saba.safe_mode == 1):
                keep_unsynchronized = 0
            sim.ri_whfast.keep_unsynchronized = keep_unsynchronized
            sim.ri_saba.keep_unsynchronized = keep_unsynchronized
            sim.integrator_synchronize()
            return sim
        else:
            if mode=='exact':
                keep_unsynchronized = 0
            if (sim.integrator=="mercurius" and sim.ri_mercurius.safe_mode == 1) or (sim.integrator=="whfast" and sim.ri_whfast.safe_mode == 1) or (sim.integrator=="saba" and sim.ri_saba.safe_mode == 1):
                keep_unsynchronized = 0
            sim.ri_whfast.keep_unsynchronized = keep_unsynchronized
            sim.ri_saba.keep_unsynchronized = keep_unsynchronized
            exact_finish_time = 1 if mode=='exact' else 0
            sim.integrate(t,exact_finish_time=exact_finish_time)
                
            return sim


    def getSimulations(self, times, **kwargs):
        """
        A generator to quickly access many simulations. 
        The arguments are the same as for `getSimulation`.
        """
        for t in times:
            yield self.getSimulation(t, **kwargs)

    
    def getBezierPaths(self,origin=None):
        """
        This function returns array that can be used as a Cubic Bezier
        Path in matplotlib. 
        The function returns two arrays, the first one contains
        the vertices for each particle and has the shape
        (Nvert, Nparticles, 2) where Nvert is the number of vertices.
        The second array returned describes the type of vertices to be
        used with matplotlib's Patch class.

        Arguments
        ---------
        origin : multiple, optional
                 If `origin` is None (default), then none of the 
                 coordinates are shifted. If `origin` is an integer
                 then the particle with that index is used as the
                 origin. if `origin` is equal to `com`, then the 
                 centre of mass is used as the origin. 


        Examples
        --------
        The following example reads in a SimulationArchive and plots
        the trajectories as Cubic Bezier Curves. It also plots the 
        actual datapoints stored in the SimulationArchive. 
        Note that the SimulationArchive needs to have enough
        datapoints to allow for smooth and reasonable orbits.

        >>> from matplotlib.path import Path
        >>> import matplotlib.patches as patches
        >>> sa = rebound.SimulationArchive("test.bin")
        >>> verts, codes = sa.getBezierPaths(origin=0)
        >>> fig, ax = plt.subplots()
        >>> for j in range(sa[0].N):
        >>>     path = Path(verts[:,j,:], codes)
        >>>     patch = patches.PathPatch(path, facecolor='none')
        >>>     ax.add_patch(patch)
        >>>     ax.scatter(verts[::3,j,0],verts[::3,j,1])
        >>> ax.set_aspect('equal')
        >>> ax.autoscale_view()
        
        """
        import numpy as np
        Npoints = len(self)*3-2
        if len(self)<=1:
            raise RuntimeError()
        Nparticles = self[0].N
        verts = np.zeros((Npoints,Nparticles,2))
        xy = np.zeros((len(self),Nparticles,2))

        if origin=="com":
            origin = -2
        elif origin is not None:
            try:
                origin = int(origin)
            except:
                raise AttributeError("Cannot parse origin")
            if origin<0 or origin>=Nparticles:
                raise AttributeError("Origin index out of range")


        for i, sim in enumerate(self):
            if origin is None:
                shift = (0,0,0,0)
            elif origin == -2:
                sp = sim.calculate_com()
                shift = (sp.x,sp.y,sp.vx,sp.vy)
            else:
                sp = sim.particles[origin]
                shift = (sp.x,sp.y,sp.vx,sp.vy)
            for j in range(sim.N):
                p = sim.particles[j]
                if i==0:
                    verts[0,j] = p.x-shift[0],p.y-shift[1]
                    verts[1,j] = p.vx-shift[2], p.vy-shift[3]
                else:
                    dt = sim.t-tlast # time since last snapshot
                    verts[-2+i*3,j] = verts[-2+i*3,j]*dt/3.+verts[-3+i*3,j]

                    verts[ 0+i*3,j] = p.x-shift[0],p.y-shift[1]

                    verts[-1+i*3,j] = -p.vx+shift[2], -p.vy+shift[3]
                    verts[-1+i*3,j] = verts[-1+i*3+0,j]*dt/3.+verts[ 0+i*3,j]

                    if i!=len(self)-1:
                        verts[+1+i*3,j] = p.vx-shift[2], p.vy-shift[3]


                xy[i,j] = p.x,p.y
            tlast = sim.t
        codes = np.full(Npoints,4,dtype=np.uint8) # Hardcoded 4 = matplotlib.path.Path.CURVE4
        codes[0] = 1 # Hardcoded 1 = matplotlib.path.Path.MOVETO
        return verts, codes



import rebound
import unittest
import math
import rebound.data
import warnings

coordinatelist = ["democraticheliocentric","whds","jacobi"]
class TestIntegratorWHFastTestParticle(unittest.TestCase):
    pass

def create_whfast_testparticle(coordinates, N, N_active):
    def do_test(self):
        sim = rebound.Simulation()
        sim.ri_whfast.coordinates = coordinates
        sim.integrator = "whfast"
        sim.dt=1e-1
        sim.add(m=1)
        for i in range(N):
            sim.add(m=0,P=1,e=0.1,f=i)
  
        sim2 = sim.copy()
        if N_active>0:
            sim2.N_active = N_active
        sim2.integrate(1)

        # some combinations can't do a simple keplerian orbit exactly (sad but true)
        eps = 1e-14
        for i in range(sim.N):
            self.assertLess(abs(sim.particles[i].x-sim2.particles[i].x),eps)
            self.assertLess(abs(sim.particles[i].vx-sim2.particles[i].vx),eps)
            self.assertLess(abs(sim.particles[i].y-sim2.particles[i].y),eps)
            self.assertLess(abs(sim.particles[i].vy-sim2.particles[i].vy),eps)
        
        # all should be able to do this exactly
        sim.integrate(1)
        eps = 1e-16
        for i in range(sim.N):
            self.assertLess(abs(sim.particles[i].x-sim2.particles[i].x),eps)
            self.assertLess(abs(sim.particles[i].vx-sim2.particles[i].vx),eps)
            self.assertLess(abs(sim.particles[i].y-sim2.particles[i].y),eps)
            self.assertLess(abs(sim.particles[i].vy-sim2.particles[i].vy),eps)
    return do_test


def create_whfast_testparticle_withplanet(coordinates, N, N_active):
    def do_test(self):
        eps = 1e-13
        sim = rebound.Simulation()
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = coordinates
        sim.dt=1e-3
        sim.add(m=1)
        sim.add(m=1e-3,P=0.4)
        sim.add(m=1e-3,P=0.7)
        for i in range(N):
            sim.add(m=0,P=1,e=0.1,f=i)
  
        sim2 = sim.copy()
        if N_active>0:
            sim2.N_active = N_active
       
        sim2.integrate(1)
        sim.integrate(1)
        for i in range(sim.N):
            self.assertLess(abs(sim.particles[i].x-sim2.particles[i].x),eps)
            self.assertLess(abs(sim.particles[i].vx-sim2.particles[i].vx),eps)
            self.assertLess(abs(sim.particles[i].y-sim2.particles[i].y),eps)
            self.assertLess(abs(sim.particles[i].vy-sim2.particles[i].vy),eps)
    return do_test

def create_whfast_testparticletype1(coordinates, N_active):
    def do_test(self):
        eps = 1e-16
        sim = rebound.Simulation()
        sim.ri_whfast.coordinates = coordinates
        sim.testparticle_type = 1
        sim.integrator = "whfast"
        sim.dt=1e-3
        sim.add(m=1)
        sim.add(m=1e-3,P=1,e=0.1)
  
        sim2 = sim.copy()
        if N_active>0:
            sim2.N_active = N_active
        sim2.integrate(1)
        sim.integrate(1)

        for i in range(sim.N):
            self.assertLess(abs(sim.particles[i].x-sim2.particles[i].x),eps)
            self.assertLess(abs(sim.particles[i].vx-sim2.particles[i].vx),eps)
            self.assertLess(abs(sim.particles[i].y-sim2.particles[i].y),eps)
            self.assertLess(abs(sim.particles[i].vy-sim2.particles[i].vy),eps)
    return do_test
def create_whfast_testparticletype1_withplanet(coordinates, N_active):
    def do_test(self):
        eps = 1e-16
        sim = rebound.Simulation()
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = coordinates
        sim.testparticle_type = 1
        sim.dt=1e-3
        sim.add(m=1)
        sim.add(m=1e-3,P=0.4)
        sim.add(m=1e-3,P=0.7,e=0.1)
        sim.add(m=1e-3,P=1.0)
  
        sim2 = sim.copy()
        if N_active>0:
            sim2.N_active = N_active
       
        sim2.integrate(1)
        sim.integrate(1)
        for i in range(sim.N):
            self.assertLess(abs(sim.particles[i].x-sim2.particles[i].x),eps)
            self.assertLess(abs(sim.particles[i].vx-sim2.particles[i].vx),eps)
            self.assertLess(abs(sim.particles[i].y-sim2.particles[i].y),eps)
            self.assertLess(abs(sim.particles[i].vy-sim2.particles[i].vy),eps)
    return do_test

## Testparticles with mass currently lead to unexpexted behaviour:
def create_whfast_massivetestparticle(coordinates, N):
    def do_test(self):
        eps = 2e-13
        sim = rebound.Simulation()
        sim.ri_whfast.coordinates = coordinates
        sim.integrator = "whfast"
        sim.dt=1e-3
        sim.add(m=1)
        sim.add(m=1e-3,P=0.4)
        for i in range(N):
            sim.add(m=0,P=1,e=0.1,f=i)
  
        sim2 = sim.copy()
        for i in range(N):
            sim2.particles[i+2].m = 1 # particles have zero mass for sim, but finite for sim2

        sim2.N_active = 2

        sim.integrate(1)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sim2.integrate(1)
            self.assertEqual(1,len(w))

        for i in range(sim.N):
            self.assertLess(abs(sim.particles[i].x-sim2.particles[i].x),eps)
            self.assertLess(abs(sim.particles[i].vx-sim2.particles[i].vx),eps)
            self.assertLess(abs(sim.particles[i].y-sim2.particles[i].y),eps)
            self.assertLess(abs(sim.particles[i].vy-sim2.particles[i].vy),eps)
    return do_test


for N in [1,2]: 
    for coordinates in coordinatelist:
        for N_active in [-1]+list(range(1,N+2)):
            test_method = create_whfast_testparticle(coordinates,N, N_active)
            test_method.__name__ = "test_whfast_testparticle_N%d_Nactive%d_"%(N,N_active)+coordinates
            setattr(TestIntegratorWHFastTestParticle, test_method.__name__, test_method)
            
        for N_active in [-1]+list(range(3,N+4)):
            test_method = create_whfast_testparticle_withplanet(coordinates,N, N_active)
            test_method.__name__ = "test_whfast_testparticle_withplanet_N%d_Nactive%d_"%(N,N_active)+coordinates
            setattr(TestIntegratorWHFastTestParticle, test_method.__name__, test_method)
        test_method = create_whfast_massivetestparticle(coordinates,N)
        test_method.__name__ = "test_whfast_massivetestparticle_N%d_"%(N)+coordinates
        setattr(TestIntegratorWHFastTestParticle, test_method.__name__, test_method)

for coordinates in coordinatelist:
    for N_active in [-1,1]:
        test_method = create_whfast_testparticletype1(coordinates, N_active)
        test_method.__name__ = "test_whfast_testparticletype1_Nactive%d_"%(N_active)+coordinates
        setattr(TestIntegratorWHFastTestParticle, test_method.__name__, test_method)
    for N_active in [-1,3]:
        test_method = create_whfast_testparticletype1_withplanet(coordinates, N_active)
        test_method.__name__ = "test_whfast_testparticletype1_withplanet_Nactive%d_"%(N_active)+coordinates
        setattr(TestIntegratorWHFastTestParticle, test_method.__name__, test_method)

if __name__ == "__main__":
    unittest.main()



import rebound
import numpy as np
import unittest
import ctypes

d = 1e-14 # precision. some orbit transformations are not well behaved

class TestOrbitalElements(unittest.TestCase):
    def test_P(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].P += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].P, os0[0].P+0.5, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].f, os0[0].f, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)
    
    def test_a(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].a += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a+0.5, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].f, os0[0].f, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)
    
    def test_e(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].e += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e+0.5, delta=d)
        self.assertAlmostEqual(os1[0].f, os0[0].f, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)
    
    def test_inc(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].inc += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].f, os0[0].f, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc+0.5, delta=d)
    
    def test_omega(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].omega += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].f, os0[0].f, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega+0.5, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)

    def test_pomega(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].pomega += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].f, os0[0].f, delta=d)
        self.assertAlmostEqual(os1[0].pomega, os0[0].pomega+0.5, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)
   
    def test_Omega(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].Omega += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].f, os0[0].f, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega+0.5, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)
    
    def test_f(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].f += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].f, os0[0].f+0.5, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)
    
    def test_M(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].M += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].M, os0[0].M+0.5, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)


    def test_l(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].l += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].l, os0[0].l+0.5, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)

    def test_theta(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].theta += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].theta, os0[0].theta+0.5, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)

    def test_T(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.2,Omega=0.3,inc=0.4,f=0.5)
        sim.add(m=1e-3,a=2)
        os0 = sim.calculate_orbits()
        sim.particles[1].T += 0.5
        os1 = sim.calculate_orbits()
        self.assertAlmostEqual(os1[0].a, os0[0].a, delta=d)
        self.assertAlmostEqual(os1[0].e, os0[0].e, delta=d)
        self.assertAlmostEqual(os1[0].T, os0[0].T+0.5, delta=d)
        self.assertAlmostEqual(os1[0].omega, os0[0].omega, delta=d)
        self.assertAlmostEqual(os1[0].Omega, os0[0].Omega, delta=d)
        self.assertAlmostEqual(os1[0].inc, os0[0].inc, delta=d)



if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
import os
import warnings

class TestSimulationArchive(unittest.TestCase):
    def test_sa_step(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.automateSimulationArchive("test.bin", step=10,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sim = rebound.Simulation("test.bin")
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
    def test_sa_fromarchive(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sim = rebound.Simulation("test.bin")
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
    def test_safe_mode_warning(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)
        x1 = sim.particles[1].x

        sa = rebound.SimulationArchive("test.bin")
        sim = sa.getSimulation(0.)
        
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sim.integrate(40.,exact_finish_time=0)
            x0 = sim.particles[1].x
            self.assertEqual(0, len(w)) # did not raise recalculating Jacobi warning

        self.assertEqual(x0,x1)
    
    def test_sa_whfasthelio_restart_safe_mode(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "democraticheliocentric"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "democraticheliocentric"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
    
    def test_sa_whfasthelio_restart(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "democraticheliocentric"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(42.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "democraticheliocentric"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
    
    def test_sa_whds_restart_safe_mode(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "whds"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "whds"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
    
    def test_sa_whds_restart(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "whds"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(42.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "whds"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
    

    def test_sa_restart_safe_mode(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 1
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
    
    def test_sa_restart(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(42.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        t1 = sim.t
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x
        t0 = sim.t

        self.assertEqual(t0,t1)
        self.assertEqual(x0,x1)

        tget = 27.123
        sim = sa.getSimulation(tget,mode="exact")
        self.assertAlmostEqual(sim.t,tget,delta=1e-14)
        tget = 25.123
        sim = sa.getSimulation(tget,mode="close")
        self.assertAlmostEqual(sim.t,tget,delta=sim.dt)

    def test_sa_restart_append(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sim.automateSimulationArchive("test.bin", 10.)
            sim.integrate(80.,exact_finish_time=0)
            self.assertEqual(1, len(w)) 

    
    def test_sa_restart_generator(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")

        times = [0.,11.,22.]
        for sim in sa.getSimulations(times,mode="close"):
            pass
        self.assertAlmostEqual(sim.t,22.058400000000105,delta=sim.dt)

    
    
    def test_sa_restart_corrector(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.ri_whfast.corrector = 5
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(42.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.ri_whfast.safe_mode = 0
        sim.ri_whfast.corrector = 5
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
        
        tget = 27.123
        sim = sa.getSimulation(tget,mode="exact")
        self.assertAlmostEqual(sim.t,tget,delta=1e-14)
        tget = 25.123
        sim = sa.getSimulation(tget,mode="close")
        self.assertAlmostEqual(sim.t,tget,delta=sim.dt)
    
    
    def test_sa_restart_ias15(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "ias15"
        sim.dt = 0.1313
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "ias15"
        sim.dt = 0.1313
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
        
        tget = 27.123
        sim = sa.getSimulation(tget,mode="exact")
        self.assertAlmostEqual(sim.t,tget,delta=1e-14)
        tget = 25.123
        sim = sa.getSimulation(tget,mode="close")
        self.assertAlmostEqual(sim.t,tget,delta=sim.dt)

#    def test_sa_restart_ias15_walltime(self):
#        sim = rebound.Simulation()
#        sim.add(m=1)
#        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
#        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
#        sim.integrator = "ias15"
#        sim.dt = 0.1313
#        sim.automateSimulationArchive("test.bin", walltime = 0.1,deletefile=True)
#        sim.integrate(3000.,exact_finish_time=0)
#
#        sim = None
#        sa = rebound.SimulationArchive("test.bin")
#        sim = sa[-1]
#        self.assertGreater(sim.t,100.)
#        sim.integrate(20000.,exact_finish_time=0)
#        x1 = sim.particles[1].x
#        
#        
#        sim = rebound.Simulation()
#        sim.add(m=1)
#        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
#        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
#        sim.integrator = "ias15"
#        sim.dt = 0.1313
#        sim.integrate(20000.,exact_finish_time=0)
#        x0 = sim.particles[1].x
#
#        self.assertEqual(x0,x1)
#        
#        tget = 116.123
#        sim = sa.getSimulation(tget,mode="exact");
#        self.assertAlmostEqual(sim.t,tget,delta=1e-14)
#        tget = 116.123
#        sim = sa.getSimulation(tget,mode="close");
#        self.assertAlmostEqual(sim.t,tget,delta=sim.dt)
#

class TestSimulationArchiveWarningsErrors(unittest.TestCase):
    def test_sa_binary_missing(self):
        with self.assertRaises(RuntimeError):
            sa = rebound.SimulationArchive("testmissing.bin")
    def test_sa_binary_version(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "ias15"
        sim.dt = 0.1313
        sim.automateSimulationArchive("test.bin", walltime = 0.01,deletefile=True)
        sim.integrate(400.,exact_finish_time=0)
        with open("test.bin","r+b") as f:
            f.seek(30)
            f.write("1.0.0     ".encode('ascii'))
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sa = rebound.SimulationArchive("test.bin")
            self.assertEqual(1, len(w)) 


class TestSimulationArchiveTmin(unittest.TestCase):
    def test_sa_tmin(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "whfast"
        sim.dt = 0.1313
        sim.integrate(400.,exact_finish_time=0)
        sim.automateSimulationArchive("test.bin", interval = 100.,deletefile=True)
        tmin = sim.t
        sim.integrate(800.,exact_finish_time=0)
        sa = rebound.SimulationArchive("test.bin")
        self.assertEqual(tmin,sa[0].t)
        self.assertEqual(tmin,sa.tmin)
        self.assertNotEqual(tmin,sa.tmax)
   

class TestSimulationArchiveMercurius(unittest.TestCase):
    def test_sa_mercurius_restart(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "mercurius"
        sim.dt = 0.1313
        sim.ri_mercurius.safe_mode = 0
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(42.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "mercurius"
        sim.dt = 0.1313
        sim.ri_mercurius.safe_mode = 0
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)
    
    def test_sa_mercurius_restart_safemode(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "mercurius"
        sim.dt = 0.1313
        sim.ri_mercurius.safe_mode = 1
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)

        sim = None
        sa = rebound.SimulationArchive("test.bin")
        sim = sa[-1]
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x
        
        
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "mercurius"
        sim.dt = 0.1313
        sim.ri_mercurius.safe_mode = 1
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        self.assertEqual(x0,x1)

    def test_append_to_corrupt_snapshot(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1e-3,a=1.)
        sim.add(m=5e-3,a=2.25)
    
        sim.automateSimulationArchive("simulationarchive.bin", interval=1000,deletefile=True) 
        sim.integrate(3001)
        with open('simulationarchive.bin', 'r+b') as f:
            f.seek(0, os.SEEK_END)          
            f.seek(f.tell() - 72, os.SEEK_SET)
            f.write(bytes(72)) # binary should be 15972 bytes, overwrite last 72 bytes with all zeros
        sa = rebound.SimulationArchive("simulationarchive.bin")
        sim = sa[-1]
        sim.automateSimulationArchive("simulationarchive.bin", interval=1000)
        sim.integrate(7001)
        sa = rebound.SimulationArchive("simulationarchive.bin")
        self.assertEqual(sa.nblobs, 8)
        self.assertAlmostEqual(sa[-1].t, 7000, places=0)
    

    def test_append_to_corrupt_snapshot_by_2bytes(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1e-3,a=1.)
        sim.add(m=5e-3,a=2.25)
    
        sim.automateSimulationArchive("simulationarchive.bin", interval=1000,deletefile=True) 
        sim.integrate(3001)
        s1 = os.path.getsize('simulationarchive.bin')
        with open('simulationarchive.bin', 'r+b') as f:
            f.seek(0, os.SEEK_END)          
            f.seek(f.tell() - 2, os.SEEK_SET)
            f.truncate() # truncate by 2 bytes
        s2 = os.path.getsize('simulationarchive.bin')
        self.assertEqual(s1, s2+2)

        sa = rebound.SimulationArchive("simulationarchive.bin")
        sim = sa[-1]
        sim.automateSimulationArchive("simulationarchive.bin", interval=1000)
        sim.integrate(7001)
        sa = rebound.SimulationArchive("simulationarchive.bin")
        self.assertEqual(sa.nblobs, 8)
        self.assertAlmostEqual(sa[-1].t, 7000, places=0)

if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
import math
import warnings

class TestParticleWarning(unittest.TestCase):
    def test_testparticle0_with_mass(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.N_active = sim.N
        sim.add(m=1,a=1)
        sim.testparticle_type = 1
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sim.integrate(1)
            self.assertEqual(0,len(w))
        sim.testparticle_type = 0
        # Warning occurs each time integrate is called
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sim.integrate(2)
            sim.integrate(3)
            self.assertEqual(2,len(w))
        sim.testparticle_hidewarnings = 1
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sim.integrate(4)
            self.assertEqual(0,len(w))


class TestParticleInSimulation(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
    
    def tearDown(self):
        self.sim = None
    
    def test_adding(self):
        self.sim.add()
        self.sim.add(m=1.)
        self.sim.add(m=1.,x=1, vy=1)
        self.sim.add(x=2)
        self.sim.add(x=3, r=1.)
        with self.assertRaises(ValueError):
            self.sim.add(x=4,a=1)
        p4 = self.sim.particles[4]
        ind = p4.index
        self.assertEqual(ind,4)

    def test_masses(self):
        self.sim.add(m=1.)
        self.sim.add(m=1.e-3, a=1.)
        self.sim.add(m=1.e-6, a=2.)
        self.sim.add(m=0., a=3.)
        ps = self.sim.particles
        self.assertAlmostEqual(ps[0].m,1.,delta=1e-15)
        self.assertAlmostEqual(ps[1].m,1.e-3,delta=1e-15)
        self.assertAlmostEqual(ps[2].m,1.e-6,delta=1e-15)
        self.assertAlmostEqual(ps[3].m,0.,delta=1e-15)

    def test_jacobi_masses(self):
        self.sim.add(m=1.)
        self.sim.add(m=1.e-3, a=1., jacobi_masses=True)
        self.sim.add(m=1.e-6, a=2., jacobi_masses=True)
        self.sim.add(m=0., a=3., jacobi_masses=True)
        ps = self.sim.particles
        self.assertAlmostEqual(ps[0].m,1.,delta=1e-15)
        self.assertAlmostEqual(ps[1].m,1.e-3,delta=1e-15)
        self.assertAlmostEqual(ps[2].m,1.e-6,delta=1e-15)
        self.assertAlmostEqual(ps[3].m,0.,delta=1e-15)
        self.assertAlmostEqual(ps[1].vy,1.000499875062461,delta=1e-15)
        self.assertAlmostEqual(ps[2].vy,0.7081066347613374,delta=1e-15)
        self.assertAlmostEqual(ps[3].vy,0.5783504759643414,delta=1e-15)
        o = self.sim.calculate_orbits(jacobi_masses=True)
        self.assertAlmostEqual(o[0].a,1.,delta=1e-15)
        self.assertAlmostEqual(o[1].a,2.,delta=1e-15)
        self.assertAlmostEqual(o[2].a,3.,delta=1e-15)
        self.assertAlmostEqual(o[0].e,0.,delta=1e-15)
        self.assertAlmostEqual(o[1].e,0.,delta=1e-15)
        self.assertAlmostEqual(o[2].e,0.,delta=1e-15)

    def test_adding_orbits(self):
        self.sim.add(m=1.)
        self.sim.add(a=1.)
        with self.assertRaises(ValueError):
            self.sim.add(e=0.1)
        with self.assertRaises(ValueError):
            self.sim.add(a=1.,e=0.1,omega=0.1,pomega=0.1)
        self.sim.add(a=2.,e=0.1,inc=0.1,pomega=0.1)
        self.sim.add(a=2.,e=0.1,inc=-0.1,pomega=0.1)
        self.sim.add(a=2.,e=0.1,inc=0.1,theta=0.1)
        self.sim.add(a=2.,e=0.1,inc=-0.1,theta=0.1)
        self.sim.add(a=2.,e=0.1,inc=0.1,l=0.1)
        self.sim.add(a=2.,e=0.1,inc=-0.1,l=0.1)
        self.sim.add(P=2.,e=0.1,inc=-2.1,pomega=0.1)
        self.sim.add(P=2.,e=0.1,inc=-2.1,pomega=0.1,f=0.2)
        self.sim.add(P=2.,e=0.1,inc=-2.1,pomega=0.1,T=0.2)
        self.sim.add(P=2.,e=0.1,inc=-2.1,pomega=0.1,theta=0.2)
        with self.assertRaises(ValueError):
            self.sim.add(a=2.,e=0.1,f=0.1,M=0.1)
        with self.assertRaises(ValueError):
            self.sim.add(a=2.,e=0.1,f=0.1,l=0.1)
        with self.assertRaises(ValueError):
            self.sim.add(a=2.,e=0.1,f=0.1,theta=0.1)
        with self.assertRaises(ValueError):
            self.sim.add(a=3.,e=1.)
        with self.assertRaises(ValueError):
            self.sim.add(a=3.,e=1.1)
        with self.assertRaises(ValueError):
            self.sim.add(a=3.,P=1.1)
        with self.assertRaises(ValueError):
            self.sim.add(a=3.,e=-0.1)
        self.sim.add(a=-3.,e=1.4)
        with self.assertRaises(ValueError):
            self.sim.add(a=-3.,e=0.9)
        self.sim.add(a=-3.,e=1.4,f=0.1)
        with self.assertRaises(ValueError):
            self.sim.add(a=-3.,e=1.4,f=3.1)
   
    def test_sim_calculate_orbits(self):
        self.sim.add(m=1.)
        self.sim.add(m=1.e-3, a=1.,e=0.2,inc=0.3)
        self.sim.add(m=1.e-3, a=2.,e=0.2,inc=0.3)
        ps = self.sim.particles
        self.assertAlmostEqual(ps[1].a,1.,delta=1e-15)
        self.assertAlmostEqual(ps[1].e,0.2,delta=1e-15)
        self.assertAlmostEqual(ps[1].inc,0.3,delta=1e-15)
        self.assertAlmostEqual(ps[2].a,2.,delta=1e-15)
        self.assertAlmostEqual(ps[2].e,0.2,delta=1e-15)
        self.assertAlmostEqual(ps[2].inc,0.3,delta=1e-15)

    def test_calculate_orbits(self):
        self.sim.add(m=1.)
        self.sim.add(a=1.)
        p = self.sim.particles
        with self.assertRaises(ValueError):
            p[0].calculate_orbit()
        o = p[1].calculate_orbit()
        string = o.__str__()
        self.assertGreater(len(string),20)
        self.assertAlmostEqual(o.a,1.,delta=1e-15)
        self.assertAlmostEqual(o.e,0.,delta=1e-15)
        self.assertAlmostEqual(o.f,0.,delta=1e-15)
        self.assertAlmostEqual(o.inc,0.,delta=1e-15)
        
        self.assertAlmostEqual(p[1].a,1.,delta=1e-15)
        self.assertAlmostEqual(p[1].e,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].f,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].inc,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].d,1.,delta=1e-15)
        self.assertAlmostEqual(p[1].v,1.,delta=1e-15)
        self.assertAlmostEqual(p[1].h,1.,delta=1e-15)
        self.assertAlmostEqual(p[1].P,math.pi*2.,delta=1e-15)
        self.assertAlmostEqual(p[1].n,1.,delta=1e-15)
        self.assertAlmostEqual(p[1].omega,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].pomega,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].Omega,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].M,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].l,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].theta,0.,delta=1e-15)
        self.assertAlmostEqual(p[1].T,0.,delta=1e-15)
    
    def test_calculate_orbits_errors(self):
        self.sim.add()
        self.sim.add(x=1)
        with self.assertRaises(ValueError):
            self.sim.particles[1].orbit

    def test_calculate_orbits_errors2(self):
        self.sim.add(m=1)
        p1 = rebound.Particle(simulation=self.sim, a=1,m=0.1)
        self.sim.add(p1)
        with self.assertRaises(ValueError):
            self.sim.particles[1].calculate_orbit(primary=p1)
    
    def test_calculate_orbits_errors3(self):
        p1 = rebound.Particle(m=1.,x=1.,vy=0.4)
        p2 = rebound.Particle(m=1.,x=4.,vy=2.4)
        with self.assertRaises(ValueError):
            p2.calculate_orbit()
        with self.assertRaises(ValueError):
            p2.calculate_orbit(primary=p1)
        p2.calculate_orbit(primary=p1,G=1.)
        

class TestParticleOperators(unittest.TestCase):
    def test_sub(self):
        p1 = rebound.Particle(m=1.1,x=1.2,vy=1.3)
        p2 = rebound.Particle(m=1.4,x=1.6,vy=1.8)
        p3 = p1 - p2
        self.assertEqual(p3.x,p1.x-p2.x)
        self.assertEqual(p3.y,p1.y-p2.y)
        self.assertEqual(p3.z,p1.z-p2.z)
        self.assertEqual(p3.vx,p1.vx-p2.vx)
        self.assertEqual(p3.vy,p1.vy-p2.vy)
        self.assertEqual(p3.vz,p1.vz-p2.vz)
        self.assertEqual(p3.m,p1.m-p2.m)

    def test_add(self):
        p1 = rebound.Particle(m=1.1,x=1.2,vy=1.3)
        p2 = rebound.Particle(m=1.4,x=1.6,vy=1.8)
        p3 = p1 + p2
        self.assertEqual(p3.x,p1.x+p2.x)
        self.assertEqual(p3.y,p1.y+p2.y)
        self.assertEqual(p3.z,p1.z+p2.z)
        self.assertEqual(p3.vx,p1.vx+p2.vx)
        self.assertEqual(p3.vy,p1.vy+p2.vy)
        self.assertEqual(p3.vz,p1.vz+p2.vz)
        self.assertEqual(p3.m,p1.m+p2.m)

    def test_mul(self):
        p1 = rebound.Particle(m=1.1,x=1.2,vy=1.3)
        p2 = 2.*p1
        self.assertEqual(p2.x,2.*p1.x)
        self.assertEqual(p2.y,2.*p1.y)
        self.assertEqual(p2.z,2.*p1.z)
        self.assertEqual(p2.vx,2.*p1.vx)
        self.assertEqual(p2.vy,2.*p1.vy)
        self.assertEqual(p2.vz,2.*p1.vz)
        self.assertEqual(p2.m,2.*p1.m)
    
    def test_div(self):
        p1 = rebound.Particle(m=1.2,x=1.4,vy=1.8)
        p2 = p1/2.
        self.assertEqual(p2.x,p1.x/2.)
        self.assertEqual(p2.y,p1.y/2.)
        self.assertEqual(p2.z,p1.z/2.)
        self.assertEqual(p2.vx,p1.vx/2.)
        self.assertEqual(p2.vy,p1.vy/2.)
        self.assertEqual(p2.vz,p1.vz/2.)
        self.assertEqual(p2.m,p1.m/2.)
    
    def test_truediv(self):
        p1 = rebound.Particle(m=1.2,x=1.4,vy=1.8)
        p2 = p1/2
        self.assertEqual(p2.x,p1.x/2.)
        self.assertEqual(p2.y,p1.y/2.)
        self.assertEqual(p2.z,p1.z/2.)
        self.assertEqual(p2.vx,p1.vx/2.)
        self.assertEqual(p2.vy,p1.vy/2.)
        self.assertEqual(p2.vz,p1.vz/2.)
        self.assertEqual(p2.m,p1.m/2.)
    
    def test_isub(self):
        p1 = rebound.Particle(m=1.1,x=1.2,vy=1.3)
        p3 = p1.copy()
        p2 = rebound.Particle(m=1.4,x=1.6,vy=1.8)
        p3 -= p2
        self.assertEqual(p3.x,p1.x-p2.x)
        self.assertEqual(p3.y,p1.y-p2.y)
        self.assertEqual(p3.z,p1.z-p2.z)
        self.assertEqual(p3.vx,p1.vx-p2.vx)
        self.assertEqual(p3.vy,p1.vy-p2.vy)
        self.assertEqual(p3.vz,p1.vz-p2.vz)
        self.assertEqual(p3.m,p1.m-p2.m)

    def test_iadd(self):
        p1 = rebound.Particle(m=1.1,x=1.2,vy=1.3)
        p3 = p1.copy()
        p2 = rebound.Particle(m=1.4,x=1.6,vy=1.8)
        p3 += p2
        self.assertEqual(p3.x,p1.x+p2.x)
        self.assertEqual(p3.y,p1.y+p2.y)
        self.assertEqual(p3.z,p1.z+p2.z)
        self.assertEqual(p3.vx,p1.vx+p2.vx)
        self.assertEqual(p3.vy,p1.vy+p2.vy)
        self.assertEqual(p3.vz,p1.vz+p2.vz)
        self.assertEqual(p3.m,p1.m+p2.m)

    def test_imul(self):
        p1 = rebound.Particle(m=1.1,x=1.2,vy=1.3)
        p2 = p1.copy()
        p2 *= 2.
        self.assertEqual(p2.x,2.*p1.x)
        self.assertEqual(p2.y,2.*p1.y)
        self.assertEqual(p2.z,2.*p1.z)
        self.assertEqual(p2.vx,2.*p1.vx)
        self.assertEqual(p2.vy,2.*p1.vy)
        self.assertEqual(p2.vz,2.*p1.vz)
        self.assertEqual(p2.m,2.*p1.m)
    
    def test_idiv(self):
        p1 = rebound.Particle(m=1.2,x=1.4,vy=1.8)
        p2 = p1.copy()
        p2 /=2.
        self.assertEqual(p2.x,p1.x/2.)
        self.assertEqual(p2.y,p1.y/2.)
        self.assertEqual(p2.z,p1.z/2.)
        self.assertEqual(p2.vx,p1.vx/2.)
        self.assertEqual(p2.vy,p1.vy/2.)
        self.assertEqual(p2.vz,p1.vz/2.)
        self.assertEqual(p2.m,p1.m/2.)
    
    def test_itruediv(self):
        p1 = rebound.Particle(m=1.2,x=1.4,vy=1.8)
        p2 = p1.copy()
        p2 /=2
        self.assertEqual(p2.x,p1.x/2.)
        self.assertEqual(p2.y,p1.y/2.)
        self.assertEqual(p2.z,p1.z/2.)
        self.assertEqual(p2.vx,p1.vx/2.)
        self.assertEqual(p2.vy,p1.vy/2.)
        self.assertEqual(p2.vz,p1.vz/2.)
        self.assertEqual(p2.m,p1.m/2.)
    


class TestParticleCopy(unittest.TestCase):
    def test_copy(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.,e=0.2,a=1)
        pc = sim.particles[1].copy()
        self.assertEqual(sim.particles[1].x,pc.x)
        self.assertEqual(sim.particles[1].vx,pc.vx)
        self.assertEqual(sim.particles[1].m,pc.m)
        sim.particles[1].m=0.01
        self.assertNotEqual(sim.particles[1].m,pc.m)
    def test_copy2(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        p1 = rebound.Particle(simulation=sim,m=1.,e=0.2,a=1)
        p2 = rebound.Particle(p1)
        p1.m=2.
        p2.m=3.
        sim.add(p1)
        sim.add(p2)
        self.assertEqual(p1.m,2.)
        self.assertEqual(p2.m,3.)
        self.assertEqual(sim.particles[1].m,2.)
        self.assertEqual(sim.particles[2].m,3.)
        

class TestParticleNotInSimulation(unittest.TestCase):
    def test_create(self):
        p1 = rebound.Particle()
        p2 = rebound.Particle(x=1)
        with self.assertRaises(ValueError):
            p3 = rebound.Particle(a=1)

if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
import ctypes


def getc(sim):
    c = []
    for i in range(sim.N):
        c.append(sim.particles[0].x)
        c.append(sim.particles[0].y)
        c.append(sim.particles[0].z)
        c.append(sim.particles[0].vx)
        c.append(sim.particles[0].vy)
        c.append(sim.particles[0].vz)
        c.append(sim.particles[0].m)
        c.append(sim.particles[0].r)
    return c

class TestTransformations(unittest.TestCase):
    
    def test_democratichelio(self):
        sim = rebound.Simulation()
        sim.add(m=1.2354)
        sim.add(m=0.1,a=1.24,e=0.123,inc=0.14,omega=0.12,Omega=0.64,l=0.632)
        sim.add(m=0.01,a=5.24,e=0.2123,inc=0.014,omega=0.012,Omega=0.0164,l=10.18632)
        sim.add(m=1e-7,a=7.24,e=0.22123,inc=0.3014,omega=0.4012,Omega=0.110164,l=2.18632)
        
        elems = (rebound.Particle * sim.N)()
        p = ctypes.cast(elems,ctypes.POINTER(rebound.Particle))

        c0 = getc(sim)

        cl = rebound.clibrebound
        cl.reb_transformations_inertial_to_democraticheliocentric_posvel(sim._particles,p,sim.N,sim.N)

        for i in range(sim.N):
            sim.particles[i].x = 1234.
            sim.particles[i].vx = 1234.
        cl.reb_transformations_democraticheliocentric_to_inertial_posvel(sim._particles,p,sim.N,sim.N)
        
        c1 = getc(sim)
        
        for i in range(len(c0)):
            self.assertAlmostEqual(c0[i],c1[i],delta=1e-16)
        
        for i in range(sim.N):
            sim.particles[i].x = 1234.
        cl.reb_transformations_democraticheliocentric_to_inertial_pos(sim._particles,p,sim.N,sim.N)
        
        c1 = getc(sim)
        
        for i in range(len(c0)):
            self.assertAlmostEqual(c0[i],c1[i],delta=1e-16)
    
    
    def test_whds(self):
        sim = rebound.Simulation()
        sim.add(m=1.2354)
        sim.add(m=0.1,a=1.24,e=0.123,inc=0.14,omega=0.12,Omega=0.64,l=0.632)
        sim.add(m=0.01,a=5.24,e=0.2123,inc=0.014,omega=0.012,Omega=0.0164,l=10.18632)
        sim.add(m=1e-7,a=7.24,e=0.22123,inc=0.3014,omega=0.4012,Omega=0.110164,l=2.18632)
        
        elems = (rebound.Particle * sim.N)()
        p = ctypes.cast(elems,ctypes.POINTER(rebound.Particle))

        c0 = getc(sim)

        cl = rebound.clibrebound
        cl.reb_transformations_inertial_to_whds_posvel(sim._particles,p,sim.N,sim.N)

        for i in range(sim.N):
            sim.particles[i].x = 1234.
            sim.particles[i].vx = 1234.
        cl.reb_transformations_whds_to_inertial_posvel(sim._particles,p,sim.N,sim.N)
        
        c1 = getc(sim)
        
        for i in range(len(c0)):
            self.assertAlmostEqual(c0[i],c1[i],delta=1e-16)
        
        for i in range(sim.N):
            sim.particles[i].x = 1234.
        cl.reb_transformations_whds_to_inertial_pos(sim._particles,p,sim.N,sim.N)
        
        c1 = getc(sim)
        
        for i in range(len(c0)):
            self.assertAlmostEqual(c0[i],c1[i],delta=1e-16)
    

    def test_jacoobi(self):
        sim = rebound.Simulation()
        sim.add(m=1.2354)
        sim.add(m=0.1,a=1.24,e=0.123,inc=0.14,omega=0.12,Omega=0.64,l=0.632)
        sim.add(m=0.01,a=5.24,e=0.2123,inc=0.014,omega=0.012,Omega=0.0164,l=10.18632)
        sim.add(m=1e-7,a=7.24,e=0.22123,inc=0.3014,omega=0.4012,Omega=0.110164,l=2.18632)
        
        elems = (rebound.Particle * sim.N)()
        p = ctypes.cast(elems,ctypes.POINTER(rebound.Particle))
        
        elemse = (ctypes.c_double * sim.N)()

        c0 = getc(sim)

        cl = rebound.clibrebound
        
        
        cl.reb_transformations_inertial_to_jacobi_posvel(sim._particles,p,sim._particles,sim.N,sim.N)

        for i in range(sim.N):
            sim.particles[i].x = 1234.
            sim.particles[i].vx = 1234.
        cl.reb_transformations_jacobi_to_inertial_posvel(sim._particles,p,sim._particles,sim.N,sim.N)
        
        c1 = getc(sim)
        
        for i in range(len(c0)):
            self.assertAlmostEqual(c0[i],c1[i],delta=1e-16)
        
        for i in range(sim.N):
            sim.particles[i].x = 1234.
        cl.reb_transformations_jacobi_to_inertial_pos(sim._particles,p,sim._particles,sim.N,sim.N)
        
        c1 = getc(sim)
        
        for i in range(len(c0)):
            self.assertAlmostEqual(c0[i],c1[i],delta=1e-16)
    
    
if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
import warnings

class TestSimulationRestartMercurius(unittest.TestCase):
    def test_sa_mercurius_restart(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "mercurius"
        sim.dt = 0.1313
        sim.ri_mercurius.safe_mode = 0
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)
        sim.save("test.bin")
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        sim = rebound.Simulation("test.bin")
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x

        self.assertEqual(x0,x1)
    
    def test_sa_mercurius_restart_safemode(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = "mercurius"
        sim.dt = 0.1313
        sim.ri_mercurius.safe_mode = 1
        sim.automateSimulationArchive("test.bin", 10.,deletefile=True)
        sim.integrate(40.,exact_finish_time=0)
        sim.save("test.bin")
        sim.integrate(80.,exact_finish_time=0)
        x0 = sim.particles[1].x

        sim = rebound.Simulation("test.bin")
        sim.integrate(80.,exact_finish_time=0)
        x1 = sim.particles[1].x

        self.assertEqual(x0,x1)
    
    
    

if __name__ == "__main__":
    unittest.main()



try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension
from codecs import open
import os
import sys

import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"

# Try to get git hash
try:
    import subprocess
    ghash = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii")
    ghash_arg = "-DGITHASH="+ghash.strip()
except:
    ghash_arg = "-DGITHASH=399189ba37c604af7bde5c42ebceabb7d34cfc7a" #GITHASHAUTOUPDATE

extra_link_args=[]
if sys.platform == 'darwin':
    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-shared')
    extra_link_args=['-Wl,-install_name,@rpath/librebound'+suffix]
    
libreboundmodule = Extension('librebound',
                    sources = [ 'src/rebound.c',
                                'src/integrator_ias15.c',
                                'src/integrator_whfast.c',
                                'src/integrator_saba.c',
                                'src/integrator_mercurius.c',
                                'src/integrator_eos.c',
                                'src/integrator_leapfrog.c',
                                'src/integrator_bs.c',
                                'src/integrator_janus.c',
                                'src/integrator_sei.c',
                                'src/integrator.c',
                                'src/gravity.c',
                                'src/boundary.c',
                                'src/display.c',
                                'src/collision.c',
                                'src/tools.c',
                                'src/derivatives.c',
                                'src/tree.c',
                                'src/particle.c',
                                'src/binarydiff.c',
                                'src/output.c',
                                'src/input.c',
                                'src/simulationarchive.c',
                                'src/transformations.c',
                                ],
                    include_dirs = ['src'],
                    define_macros=[ ('LIBREBOUND', None) ],
                    # Removed '-march=native' for now.
                    extra_compile_args=['-fstrict-aliasing', '-O3','-std=c99','-Wno-unknown-pragmas', ghash_arg, '-DLIBREBOUND', '-D_GNU_SOURCE', '-fPIC'],
                    extra_link_args=extra_link_args,
                    )

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='rebound',
    version='3.19.4',
    description='An open-source multi-purpose N-body code',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/hannorein/rebound/',
    author='Hanno Rein',
    author_email='hanno@hanno-rein.de',
    license='GPL',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Scientific/Engineering :: Astronomy',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords='astronomy astrophysics nbody integrator symplectic wisdom-holman',
    packages=['rebound'],
    package_data = {'rebound':['rebound.h']},
    install_requires=[],
    tests_require=["numpy","matplotlib"],
    test_suite="rebound.tests",
    ext_modules = [libreboundmodule],
    zip_safe=False)



import rebound
import unittest
import math 
from ctypes import c_uint32

class TestHash(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(a=0.1)
        self.sim.add(a=1.)
        self.sim.add(a=5.)
        self.sim.particles[2].hash = "earth"
        self.sim.particles[3].hash = "jupiter"

    def tearDown(self):
        self.sim = None

    def test_rebound_hash(self):
        self.assertNotEqual(rebound.hash("foo").value, rebound.hash("bar").value)
        self.assertEqual(rebound.hash("earth").value, 1424801690)
        
    def test_add_hash(self):
        self.assertEqual(self.sim.particles[0].hash.value, 0)
        self.assertEqual(self.sim.particles[2].hash.value, 1424801690)

    def test_particles(self):
        self.assertAlmostEqual(self.sim.particles["earth"].a, 1., delta=1e-15)
        self.assertAlmostEqual(self.sim.particles["jupiter"].a, 5., delta=1e-15)
        with self.assertRaises(rebound.ParticleNotFound):
            self.sim.particles["venus"]

    def test_removing_particles(self):
        self.sim.add(a=30.)
        self.sim.remove(0)
        self.sim.remove(0)
        self.sim.remove(hash=rebound.hash("earth"), keepSorted=0)
        self.assertEqual(self.sim.particles["jupiter"].hash.value, rebound.hash("jupiter").value)
        self.assertEqual(self.sim.N, 2)
        with self.assertRaises(rebound.ParticleNotFound):
            self.sim.particles["earth"]
        self.sim.remove(hash="jupiter")
        with self.assertRaises(rebound.ParticleNotFound):
            self.sim.particles["jupiter"]

    def test_adding_particles(self):
        self.assertAlmostEqual(self.sim.particles["earth"].a, 1., delta=1e-15)
        self.assertEqual(self.sim.N_lookup, 3)
        self.sim.add(a=10.)
        self.sim.particles[4].hash = "saturn"
        self.assertAlmostEqual(self.sim.particles["jupiter"].a, 5., delta=1e-15)
        self.assertAlmostEqual(self.sim.particles["saturn"].a, 10., delta=1e-15)
        self.assertAlmostEqual(self.sim.particles["earth"].a, 1., delta=1e-15)
        self.assertEqual(self.sim.N_lookup, 4)

    def test_add(self):
        self.sim.add(a=7., hash = "planet 9")
        self.assertEqual(rebound.hash("planet 9").value, self.sim.particles["planet 9"].hash.value)

class TestEmptyLookup(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.add(m=1.)
        self.sim.add(a=0.1)
        self.sim.add(a=1.)
        self.sim.add(a=5.)

    def tearDown(self):
        self.sim = None

    def test_empty_lookup(self):
        with self.assertRaises(rebound.ParticleNotFound):
            self.sim.particles["venus"]

class TestZeroHash(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()

    def tearDown(self):
        self.sim = None

    def test_zero_first_hash(self):
        self.sim.add(m=1., hash=c_uint32(0))
        self.sim.add(m=2., hash=c_uint32(1))
        self.assertAlmostEqual(self.sim.particles[c_uint32(1)].m, 2., delta=1e-15)
        self.assertEqual(self.sim.N_lookup, 2)
        self.sim.add(m=3., hash="jupiter")
        self.assertAlmostEqual(self.sim.particles[c_uint32(0)].m, 1., delta=1e-15)
        self.assertAlmostEqual(self.sim.particles["jupiter"].m, 3., delta=1e-15)
        self.assertEqual(self.sim.N_lookup, 3)
    
    def test_zero_second_hash(self):
        self.sim.add(m=1., hash=c_uint32(1))
        self.sim.add(m=2., hash=c_uint32(0))
        self.assertAlmostEqual(self.sim.particles[c_uint32(0)].m, 2., delta=1e-15)
        self.assertEqual(self.sim.N_lookup, 2)
        self.sim.add(m=3., hash="jupiter")
        self.assertAlmostEqual(self.sim.particles[c_uint32(0)].m, 2., delta=1e-15)
        self.assertAlmostEqual(self.sim.particles["jupiter"].m, 3., delta=1e-15)
        self.assertEqual(self.sim.N_lookup, 3)

class TestSort(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()

    def tearDown(self):
        self.sim = None

    def init(self, hashes):
        for hash in hashes:
            if hash is None:
                self.sim.add(m=0.)
            else:
                self.sim.add(m=hash, hash=hash)
   
    def case(self, hashes):
        self.setUp()
        self.init(hashes)
        for hash in hashes:
            if hash is not None:
                self.assertAlmostEqual(self.sim.particles[c_uint32(hash)].m, hash, delta=1e-15)
        self.tearDown()

    def test_hash(self):
        self.case(hashes=[0,1,2,3])
        self.case(hashes=[0,None,2,3])
        self.case(hashes=[2,3,1,4])
        self.case(hashes=[2,4,None,3])
        self.case(hashes=[2,0,4,3])
        self.case(hashes=[2,0,4,None])
        self.case(hashes=[2,3,4,0])
        self.case(hashes=[None,4,3,0])
        e = rebound.hash("earth").value
        self.case(hashes=[0,1,2,e])
        self.case(hashes=[0,None,e,3])
        self.case(hashes=[2,3,e,4])
        self.case(hashes=[2,e,None,3])
        self.case(hashes=[2,0,e,3])
        self.case(hashes=[e,0,4,None])
        self.case(hashes=[2,e,4,0])
        self.case(hashes=[None,4,e,0])

if __name__ == "__main__":
    unittest.main()
